package field

import (
	"context"
	"math"
	"reflect"
	"testing"
	"time"

	"git.perx.ru/perxis/perxis-go/pkg/errors"
	"github.com/stretchr/testify/assert"
	"github.com/stretchr/testify/require"
)

type CustomInt int

func TestNumberField_Decode(t *testing.T) {
	tests := []struct {
		name    string
		field   *Field
		data    interface{}
		want    interface{}
		wantErr bool
	}{
		{"Correct", Number("int"), int64(2), int64(2), false},                               // #0
		{"Correct", Number("int"), 2.2, int64(2), false},                                    // #1
		{"Correct", Number("int"), 2, int64(2), false},                                      // #2
		{"Correct", Number("int"), float32(2.2), int64(2), false},                           // #3
		{"Correct", Number("int"), float64(2.6), int64(3), false},                           // #4
		{"Correct", Number("int"), 2.6, int64(3), false},                                    // #5
		{"Correct", Number("int"), CustomInt(2), int64(2), false},                           // #5
		{"MaxInt64", Number(NumberFormatInt), int64(math.MaxInt64), nil, true},              // #6
		{"MinInt64", Number(NumberFormatInt), int64(math.MinInt64), nil, true},              // #7
		{"maxInt in float", Number(NumberFormatInt), float64(maxInt), int64(maxInt), false}, // #8
		{"minInt in float", Number(NumberFormatInt), float64(minInt), int64(minInt), false}, // #9
		{"Convert error", Number(NumberFormatInt), math.MaxFloat64, nil, true},              // #10
		{"Convert error", Number(NumberFormatInt), -math.MaxFloat64, nil, true},             // #11
		{"Convert error", Number(NumberFormatInt), float64(math.MaxInt64), nil, true},       // #13
		{"Convert error", Number(NumberFormatInt), float64(math.MinInt64), nil, true},       // #14

		{"Correct", Number("float"), int8(2), 2.0, false},  // #15
		{"Correct", Number("float"), 2.2, 2.2, false},      // #16
		{"Correct", Number("float"), 2, 2.0, false},        // #17
		{"Correct", Number("float"), int64(2), 2.0, false}, // #18

		{"Wrong data", Number("int"), "", nil, true},         // #0
		{"Wrong data", Number("int"), []byte(""), nil, true}, // #1

		{"Wrong data", Number("float"), "", nil, true},         // #2
		{"Wrong data", Number("float"), []byte(""), nil, true}, // #3
	}
	for _, tt := range tests {
		t.Run(tt.name, func(t *testing.T) {
			got, err := Decode(context.Background(), tt.field, tt.data)
			if (err != nil) != tt.wantErr {
				t.Errorf("Decode() error = %v, wantErr %v", err, tt.wantErr)
				return
			}
			if !reflect.DeepEqual(got, tt.want) {
				t.Errorf("Decode() got = %v, want %v", got, tt.want)
			}
		})
	}
}

func TestNumberField_Encode(t *testing.T) {
	tests := []struct {
		name    string
		field   *Field
		data    interface{}
		want    interface{}
		wantErr bool
	}{
		{"Correct", Number("int"), int64(2), int64(2), false},     // #0
		{"Correct", Number("int"), 2.2, int64(2), false},          // #1
		{"Correct", Number("int"), 2, int64(2), false},            // #2
		{"Correct", Number("int"), float32(2.2), int64(2), false}, // #3
		{"Correct", Number("int"), float32(2.6), int64(3), false}, // #4
		{"Correct", Number("int"), 2.6, int64(3), false},          // #5

		{"Correct", Number("float"), int8(2), 2.0, false},  // #6
		{"Correct", Number("float"), 2.2, 2.2, false},      // #7
		{"Correct", Number("float"), 2, 2.0, false},        // #8
		{"Correct", Number("float"), int64(2), 2.0, false}, // #9

		{"Wrong data", Number("int"), "", nil, true},         // #0
		{"Wrong data", Number("int"), []byte(""), nil, true}, // #1

		{"Wrong data", Number("float"), "", nil, true},         // #2
		{"Wrong data", Number("float"), []byte(""), nil, true}, // #3
	}
	for _, tt := range tests {
		t.Run(tt.name, func(t *testing.T) {
			got, err := Encode(context.Background(), tt.field, tt.data)
			if (err != nil) != tt.wantErr {
				t.Errorf("Decode() error = %v, wantErr %v", err, tt.wantErr)
				return
			}
			if !reflect.DeepEqual(got, tt.want) {
				t.Errorf("Decode() got = %v, want %v", got, tt.want)
			}
		})
	}
}

func TestNumberField_DecodeNonStrict(t *testing.T) {
	tests := []struct {
		name    string
		field   *Field
		data    interface{}
		want    interface{}
		wantErr string
	}{
		{
			name:  "Nil data",
			field: Number(NumberFormatInt),
			data:  nil,
			want:  nil,
		},
		{
			name:  "Correct int",
			field: Number(NumberFormatInt),
			data:  int64(2),
			want:  int64(2),
		},
		{
			name:  "Correct float",
			field: Number(NumberFormatFloat),
			data:  2.2,
			want:  2.2,
		},
		{
			name:  "String as float",
			field: Number(NumberFormatFloat),
			data:  "2.2",
			want:  2.2,
		},
		{
			name:  "String as int",
			field: Number(NumberFormatInt),
			data:  "2",
			want:  int64(2),
		},
		{
			name:    "String incorrect",
			field:   Number(NumberFormatInt),
			data:    "string",
			wantErr: "error convert \"string\" to number",
		},
		{
			name:    "Bool",
			field:   Number(NumberFormatInt),
			data:    true,
			wantErr: "decode error: error convert \"%!s(bool=true)\" to number",
		},
		{
			name:    "Time",
			field:   Number(NumberFormatInt),
			data:    time.Date(2023, 10, 1, 12, 0, 0, 0, time.UTC),
			wantErr: "error convert \"2023-10-01 12:00:00 +0000 UTC\" to number",
		},
		{
			name:    "Array",
			field:   Number(NumberFormatInt),
			data:    []interface{}{1, "2", nil},
			wantErr: "error convert \"[%!s(int=1) 2 <nil>]\" to number",
		},
		{
			name:    "Object",
			field:   Number(NumberFormatInt),
			data:    map[string]any{"key": "value"},
			wantErr: "error convert \"map[key:value]\" to number",
		},
	}
	for _, tt := range tests {
		t.Run(tt.name, func(t *testing.T) {
			got, err := Decode(context.Background(), tt.field, tt.data, NonStrictMode())
			if tt.wantErr != "" {
				require.Error(t, err)
				assert.ErrorContains(t, err, tt.wantErr)
				return
			}

			require.NoError(t, err)
			assert.Equal(t, tt.want, got)
		})
	}
}

// Добавлен полные функции, чтобы результаты теста не поменялись, в случае изменения Decode
// Реализация cast type.
func toNumber(i interface{}) (interface{}, error) {
	switch v := i.(type) {
	case int64:
		return v, nil
	case int:
		return int64(v), nil
	case int8:
		return int64(v), nil
	case int32:
		return int64(v), nil
	case uint64:
		return v, nil
	case uint:
		return uint64(v), nil
	case uint8:
		return uint64(v), nil
	case uint32:
		return uint64(v), nil
	case float32:
		return float64(v), nil
	case float64:
		return v, nil
	}
	return 0, errors.Errorf("error convert \"%s\" to number", i)
}
func decodeNumber(field *Field, v interface{}) (interface{}, error) {
	params, ok := field.Params.(*NumberParameters)
	if !ok {
		return nil, errors.New("field parameters required")
	}

	if v == nil {
		return v, nil
	}

	n, err := toNumber(v)
	if err != nil {
		return nil, err
	}

	switch params.Format {
	case NumberFormatInt:
		switch i := n.(type) {
		case int64:
			if i > maxInt || i < minInt {
				return nil, errors.New("integer out of range")
			}
			return i, nil
		case uint64:
			if i > maxInt {
				return nil, errors.New("integer out of range")
			}
			return i, nil
		case float64:
			if i > maxInt || i < minInt {
				return nil, errors.New("integer out of range")
			}
			return int64(math.Round(i)), nil
		}
	case NumberFormatFloat:
		switch i := n.(type) {
		case float64:
			return i, nil
		case int64:
			return float64(i), nil
		case uint64:
			return float64(i), nil
		}
	}
	return n, nil
}

// Реализация reflect.
func decodeNumberReflect(field *Field, v interface{}) (interface{}, error) { //nolint:unparam // test case
	params, ok := field.Params.(*NumberParameters)
	if !ok {
		return nil, errors.New("field parameters required")
	}

	if v == nil {
		return v, nil
	}

	n, err := toNumberReflect(v) // вызов метода reflect
	if err != nil {
		return nil, err
	}

	switch params.Format {
	case NumberFormatInt:
		switch i := n.(type) {
		case int64:
			if i > maxInt || i < minInt {
				return nil, errors.New("integer out of range")
			}
			return i, nil
		case uint64:
			if i > maxInt {
				return nil, errors.New("integer out of range")
			}
			return i, nil
		case float64:
			if i > maxInt || i < minInt {
				return nil, errors.New("integer out of range")
			}
			return int64(math.Round(i)), nil
		}
	case NumberFormatFloat:
		switch i := n.(type) {
		case float64:
			return i, nil
		case int64:
			return float64(i), nil
		case uint64:
			return float64(i), nil
		}
	}
	return n, nil
}

// команда запуска go test -bench=.
func BenchmarkDecodeReflectNumber(b *testing.B) {
	for i := 0; i < b.N; i++ {
		_, _ = decodeNumberReflect(Number(NumberFormatInt), i)
	}
}
func BenchmarkDecodeReflectCustomNumber(b *testing.B) {
	for i := 0; i < b.N; i++ {
		_, _ = decodeNumberReflect(Number(NumberFormatInt), customInt(i))
	}
}
func BenchmarkDecodeNumber(b *testing.B) {
	for i := 0; i < b.N; i++ {
		_, _ = decodeNumber(Number(NumberFormatInt), i)
	}
}
func BenchmarkDecodeNumberCurrentSolution(b *testing.B) {
	for i := 0; i < b.N; i++ {
		_, _ = NumberType{}.decode(context.Background(), Number(NumberFormatInt), i)
	}
}
func BenchmarkDecodeCustomNumberCurrentSolution(b *testing.B) {
	for i := 0; i < b.N; i++ {
		_, _ = NumberType{}.decode(context.Background(), Number(NumberFormatInt), customInt(i))
	}
}
